#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "badpenguin-ansi.h"
#include "badpenguin-chain.h"
#include "badpenguin-errors.h"
#include "badpenguin-fs.h"
#include "badpenguin-untar.h"
#include "gnutar.h"
#include "untar-blockcache.h"
#include "untar-namecache.h"
#include "untar-support.h"


// Handle a cache of block containing information on the owner
// and permission to set on the file system, permissions are
// handled from bottom to top
TChain *blockcache = NULL;


// Insert a new block into the double linked list of blocks
int push_block( char *target_file, TTarBlock *buffer ) {
  TChainItem  *newitem;
  TCacheBlock *newBlock;
  if (blockcache==NULL) blockcache = chain_create();
  newitem = chain_append(blockcache);
  newBlock = (TCacheBlock*) calloc( 1, sizeof(TCacheBlock) );
  newitem->data = newBlock;
  // Decode values
  newBlock->name = (char*) strdup(target_file);
  newBlock->mode = oct2int( buffer->mode, 8 );
  newBlock->uid = lookup_user( buffer->uname, buffer );
  newBlock->gid = lookup_group( buffer->gname, buffer );
  newBlock->mtime = oct2int( buffer->mtime, 12 );
  newBlock->type  = buffer->typeflag;
  return 0;
}


// Flush the block cache applying permissions to the FS from bottom to top
int FlushCache() {
  TChainItem  *item = NULL;
  TCacheBlock *block = NULL;
  int retval = 0;
  
  if ( blockcache == NULL ) {
    printWARNING("untar: flushcache has no blocks to flush\n");
    normal_color();
    return 0;
  }
  if (giVerbose>1) printf("- flushing block cache \n");

  item = blockcache->tail;
  while (item!=NULL) {
    // Apply values
    // set the utime before chmod because some versions of utime are
    // broken and trash the modes of the file.
    block = item->data;
    //printf("- untar_FlushCache: %s\n", lastBlock->name );    

    // DO NOT CHANGE TIME AND MODE FOR SYMLINK
    if (block->type!=TAR_SYMTYPE) {
      // Change A. Time      
      if ( 0 != file_time( block->name, block->mtime, gsTime ) ) {
        printERROR("untar: errore in file_time %s\n", block->name );
        normal_color();
        retval = ERROR_IO;
      }
      // Change Mode
      if ( 0 != chmod( block->name, block->mode ) ) {
        printERROR("untar: errore in chmod %s\n", block->name );
        normal_color();
        retval = ERROR_IO;
      }
    }
    // Change Owner
    if ( 0 != chown( block->name, block->uid, block->gid ) ) {
      printERROR("untar: errore in chown %s\n", block->name );
      normal_color();
      retval = ERROR_IO;
    }
    // Go previous and free
    item = item->prev;
  }
  chain_cleanall(blockcache);
  return retval;
}
