#include <sys/stat.h>
#include <sys/sysmacros.h>

#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <zlib.h>

#include "badpenguin-ansi.h"
#include "badpenguin-errors.h"
#include "badpenguin-fs.h"
#include "badpenguin-untar.h"
#include "gnutar.h"
#include "untar-blockcache.h"
#include "untar-support.h"

// Global Options
char	   *gsRootDir = NULL;   // lpopt->targetdir
char	   *gsTargetDir = NULL;   // lpopt->targetdir + lpopt->prefixdir
int	    giAction  = TAR_ACTION_UNKNOWN;
int	    giMode    = 0;	// untar(...mod)
int         giVerbose = 0;      // Verbose Level
time_t      gsTime;		// Time of the beggining of the installation

// Gzipped archive file related vars
long        gzblock  = -1;	// Nr. of readed block from the zipped stream
TTarBlock   gzbuffer;		// buffer with the current readed data block
int         gzcompleted = 0;	// 1=untar completed
gzFile     *gzfd = NULL;	// Pointer to the compressed files
int         gzheader = 1;	// 1=reading header - 0=data block
int	    loFlagInstall = 0;	// Flag 0=Normal, 1=trovato /install, 2=fine /install

// Current Extraced file related vars
char        lo_source_file[256];
time_t      lo_source_time;
long        lo_source_size;
char        lo_target_file[256];
FILE       *lo_target_fd = NULL;  // Handle del file target da creare
long        lo_target_tocopy = 0; // Nr. di byte rimasti da copiare su target




/*
  STRATEGY
  To avoid problem with files and directory permissions we first extract
  all the files with mode 0700 and uid/gid root.root, then every entry
  is cached. At the end of the extraction process we apply all the
  cached entries into reverse order.
*/



// ===============================================
// Skip n block for 'size' bytes into the tar file
// ===============================================

int goto_next_header(gzFile *fd, long size) {
  // Byte to skip to reach the other header
  z_off_t skip_byte;
  int     skip_block;
  if ( size > 0 ) {
    // Skip data Blocks
    skip_block = ( size / TAR_BLOCKSIZE );
    if ( size % TAR_BLOCKSIZE != 0 ) {
      skip_block++;
    }
    // Legge i blocchi dal file compresso
    skip_byte = skip_block * TAR_BLOCKSIZE;
    if ( -1 == gzseek( fd, skip_byte, SEEK_CUR ) ) {
      printERROR("untar: errore durante il posizionamento al blocco successivo a %li\n", gzblock );
      normal_color();
      return ERROR_IO;
    }
    // Avanza gzblock del nr. di blocchi da spostare
    gzblock += skip_block;
  }
  return 0;
}


/* 
 * Controlla l'operazione da effettuare in accordo con l'azione 
 * specificata. Se il file/directory passa il test viene creato
 * e la sua entry viene cacheata per essere in seguito elaborata
 * per applicare le giuste permissions.
 */
int process_file() {
  int     lo_target_exist;      // Flag, 0 se il target file gia' esiste
  struct  stat lo_stat;         // Struttura per controllare se un file esiste
  mode_t  lo_flag_block;
  int     lo_flag_makedir;
  int     lo_major, lo_minor;   // Major and minor of the device to create
  
  if (giVerbose>2) printf( "- %s (%li) ==> %s \n", lo_source_file, lo_source_size, lo_target_file );

  switch(giAction) {
      
    case TAR_ACTION_INSTALL:
    case TAR_ACTION_UPDATE:
    case TAR_ACTION_RESTORE:
    {

      // Controlla se il file target esiste e se bisogna installarlo o meno
      lo_target_exist = lstat( lo_target_file, &lo_stat );
      if ( lo_target_exist == 0 ) {
        if (giVerbose>2) printf("- %s exists\n", lo_target_file );
        // Nel caso esiste, deve controllare la data ?
        if (giAction == TAR_ACTION_UPDATE) {
          if ( lo_stat.st_mtime == lo_source_time ) {
            if (giVerbose>0) printf("- file %s already updated, skipping\n", lo_target_file);
            goto_next_header( gzfd, lo_source_size );
            return 0;
          } else {
            if (giVerbose>0) printf("> file %s is out of date, updating\n", lo_target_file);
          }
        } else
        if (giAction == TAR_ACTION_RESTORE) {
          // File exist so skip it
          if (giVerbose>0) printf("- file %s already exists, skipping\n", lo_target_file);
          goto_next_header( gzfd, lo_source_size );
          return 0;
        }
      }

      // Il file e' da installare: procedi
      switch (gzbuffer.typeflag) {

        // File Normale
        case TAR_REGTYPE:
        case TAR_AREGTYPE: 
	case TAR_CONTTYPE:
        {
          // Se il file esiste deve essere prima rimosso
          if (lo_target_exist == 0 ) unlink( lo_target_file );
          // Crea il file
          lo_target_fd = fopen( lo_target_file, "w");
          if ( lo_target_fd == NULL) {
            /* Impossibile creare il file, probabilmente non esiste
             * la directory in cui deve essere creato. Quindi proviamo
             * a creare la directory ed a riprovare la creazione del
             * file
             */
	    printWARNING("untar: errore nella creazione del file %s, provo a creare la directory padre\n", lo_target_file); 
            normal_color();
            makedir( 
              dirpath( lo_target_file ), 
              S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH 
            );
            lo_target_fd = fopen(lo_target_file,"w");
            if ( lo_target_fd == NULL) {
              lo_target_fd = NULL;
              printERROR("untar: impossibile creare il file %s\n", lo_target_file );
              normal_color();
              //goto_next_header( gzfd, lo_source_size );
              return ERROR_IO;
            }
          }
          /*
           * Il file e' stato creato, se la dimensione e' maggiore di
           * zero i blocchi successivi sono blocchi dati. Invice se
           * la dimensione e' 0 il prossimo blocco e' un'altra header
           * quindi dobbiamo creare un file di lunghezza 0
           */
          if (lo_source_size > 0 ) {
            lo_target_tocopy = lo_source_size;
            gzheader = 0;  // Get data Block
            if (giVerbose>1) printf("> creazione file normale %s\n", lo_target_file); 
          } else {
            fclose( lo_target_fd );
            lo_target_fd = NULL;
            gzheader = 1;
            if (giVerbose>1) {
              printWARNING("> creazione file nullo %s \n", lo_target_file);
              normal_color();
	    }
          }
          push_block( lo_target_file, &gzbuffer );
	  return 0;
	  // Fine file normale
        }

        // Hard Link
        case TAR_LNKTYPE: {
          if (giVerbose>1) printf("> hard link %s \n", lo_target_file );
          // Rimuove il file precedete se esiste
          if (lo_target_exist == 0 ) unlink( lo_target_file );
          /*
           * In caso di problemi, probabilmente la disposizione dei FS
           * e' diversa, quindi proviamo a creare un link simbolico
           * invece.
           */           
          if ( 0 != link( gzbuffer.linkname, lo_target_file ) ) {
            printWARNING("untar: errore nella creazione hard-linking su %s, provo con un sym-link\n", gzbuffer.linkname );
            normal_color();
            if ( 0 != symlink( gzbuffer.linkname, lo_target_file ) ) {
              printERROR("untar: impossibile creare anche il symlink\n" );
              normal_color();
              return ERROR_IO;
            }
          }
          push_block( lo_target_file, &gzbuffer );
	  return 0;
	  // Fine Hard Link
        }

        // Lynk simbolico
        case TAR_SYMTYPE: {
          // Rimuove il precedente file se esiste
          if (lo_target_exist == 0 ) unlink( lo_target_file );
          if (giVerbose>1) printf("> creazione symlink su [%s]\n", gzbuffer.linkname );
          if ( 0 != symlink(gzbuffer.linkname, lo_target_file) ) {
            printERROR("untar: errore creazione del symlink %s --> %s\n", lo_target_file, gzbuffer.linkname );
            normal_color();
            return ERROR_IO;
          }
          push_block( lo_target_file, &gzbuffer );
          return 0;
	  // fine symlink
        }

        // File di Device
        case TAR_CHRTYPE:
        case TAR_BLKTYPE: {
          // Remove first
          if (lo_target_exist == 0 ) unlink( lo_target_file );
          if (giVerbose>1) printf("> creazione del file di device %s\n", lo_target_file);
          if ( gzbuffer.typeflag == TAR_CHRTYPE ) {
            lo_flag_block = S_IFCHR;
          } else {
            lo_flag_block = S_IFBLK;
          }
	  lo_major = oct2int( gzbuffer.devmajor, 8);
	  lo_minor = oct2int( gzbuffer.devminor, 8);
          if ( 0 != mknod( lo_target_file, lo_flag_block, makedev( lo_major, lo_minor) ) ) {
            printERROR("untar: errore creazione device %s\n", lo_target_file );
            normal_color();
            return ERROR_IO;
	  }
          push_block( lo_target_file, &gzbuffer );
          return 0;
	  // fine device
	}

        // Directory
        case TAR_DIRTYPE: {
          if (giVerbose>1) printf("> creazione directory %s\n", lo_target_file);
          if (lo_target_exist == 0 ) {
            //printf("	provo a rimuovere la directory\n");
            lo_flag_makedir = rmdir( lo_target_file );
          } else {
            lo_flag_makedir = 0;
          }
          if ( lo_flag_makedir == 0 ) {
            //printf("	creazione della directory 0700\n");
            if ( 0 != mkdir( lo_target_file,  S_IRWXU ) ) {
              printERROR("untar: errore nella creazione della directory %s", lo_target_file );
              normal_color();
	      return ERROR_IO;
            }
          } else {
            if (giVerbose>1) printf("> creazione dir %s non necessaria\n", lo_target_file);
          }
          push_block( lo_target_file, &gzbuffer );
          return 0;
	  // fine dir
        }

        // Fifo             
        case TAR_FIFOTYPE: {
          // Remove first
          if (lo_target_exist == 0 ) unlink( lo_target_file );
          if (giVerbose>1) printf("> creazione fifo %s\n", lo_target_file);
          if ( 0 == mkfifo( lo_target_file, S_IRWXU ) ) {
            push_block( lo_target_file, &gzbuffer );
          } else {
            printERROR( "untar: errore durante la creazione della FIFO %s\n", lo_target_file);
            normal_color();
	    return ERROR_IO;
          }
          return 0;
	  // fine fifo
        }

        // File di tipo sconosciuto
        default: {
          printERROR( "untar: blocco %li [%s] e' di tipo sconosciuto (%c) !\n", gzblock, lo_source_file, gzbuffer.typeflag );
	  normal_color();
	  return ERROR_PARAMETER;
        }            
      // Fine switch tipo di file
      }
    // fine case su action
    }
    
    // default case for not handled yet actions
    default: {
      printERROR("untar: l'azione scelta (%i) non e' ancora supportata \n", giAction); 
      normal_color();
      return ERROR_PARAMETER;
    }
  // fine switch su Action
  }
}



/* ========================================================================
 * Check for the Block, if it is an header parse the information in it
 * else if it is a data bloack append it to the queue of the last open file
 * ========================================================================
 */
 
int process_block() {

  // Elabora il blocco come header
  if ( gzheader == 1 ) {

    // Controlla che non sia un blocco terminatore    
    if (( gzbuffer.name[0] == 0 ) && ( gzbuffer.magic[0] == 0 )) {
      if (giVerbose>0) {
	printSUCCESS( "- estrazione archivio completata\n");
        normal_color();
      }
      gzcompleted=1;
      return 0;
    }

    // Controlla la corretta Intestazione del blocco
    if ( strncmp( gzbuffer.magic, TAR_MAGIC, 4 ) != 0 ) {
      printERROR( "untar: block nr. %li is not an header Block!\n", gzblock );
      normal_color();
      sleep(1);
      return ERROR_IO;
    }

    // SPECIAL: TAR_LONGLINK
    if (gzbuffer.typeflag == TAR_LONGLINK) {
      if (giVerbose>0) printf( "untar: block nr. %li e' un LONGLINK !\n", gzblock );
      // Legge il nome dal successivo blocco di dati non compressi
      gzblock++;
      gzread( gzfd, (char*)(&gzbuffer), TAR_BLOCKSIZE );
      // Assegna il long name alla variabile
      strcpy(lo_source_file, (char*)(&gzbuffer) );
      if (giVerbose>1) printf("- Link Name (%i) = %s \n", strlen(lo_source_file), lo_source_file );
      // Legge il blocco dati con le informazioni
      gzblock++;
      gzread( gzfd, (char*)(&gzbuffer), TAR_BLOCKSIZE );
      //
      if (giVerbose>1) printf("- reset flag (%c)(%i)\n", gzbuffer.typeflag, (int)(gzbuffer.typeflag) );

    } else {

      // ESCLUDI

      /* 
       * Aggiunge al nome del file sorgente l'eventuale prefisso,
       * quindi compone il nome del file target aggiungendo la directory
       * d'installazione al nome del file.
       */
      // Se il file non ha prefisso usa il solo nome del file
      if ( gzbuffer.prefix[0] == 0 ) {
        strcpy( lo_source_file, gzbuffer.name );
      } else {
        // ... altrimenti mette il prefisso prima del nome
        strcpy( lo_source_file, gzbuffer.prefix);
        if ( lo_source_file[strlen(lo_source_file)] != '/' ) {
          strcat( lo_source_file, "/" );
        }
        strcat( lo_source_file, gzbuffer.name );
      }
    }
    
    // Non contento aggiunge il nome della target directory
    strcpy( lo_target_file, gsTargetDir );
    strcat( lo_target_file, "/" );
    strcat( lo_target_file, lo_source_file );
    // Converte i valori di dimensione e data modifica
    lo_source_time = (time_t) oct2int( gzbuffer.mtime, 12 );
    lo_source_size = oct2int( gzbuffer.size, 12);

    // Controlla la maschera del file /install
    if ( giMode == TAR_MODE_HEADER ) {
      if ( 0 != strncmp( "install/", lo_source_file, 8 ) ) {
        //printf("- Salto il file non in INSTALL %s\n", lo_source_file);
	#ifdef BROKEN
	if (loFlagInstall == 1 ) {
	  // Non ci sono piu' /install, termina pure la lettura dell'archivio
	  gzcompleted = 1;
	  return 0;
	}
	#endif
        goto_next_header( gzfd, lo_source_size );	  
	return 0;
      }
      // Allora elabora il file, marchiamo che abbiamo trovato install
      loFlagInstall = 1;
      return process_file();    
    }

    // Skip install files upon installation
    if ( giMode == TAR_MODE_ALL ) {
      if ( 0 == strncmp( "install/", lo_source_file, 8 ) ) {
        goto_next_header( gzfd, lo_source_size );	  
	return 0;      
      }
    }
    
    // Controlla la maschera dei file nel caso incoming
    if ( 0 == strncmp( "incoming/", lo_source_file, 9 ) ) {
      //TODO: aggiustare INCOMING
      strcpy( lo_source_file, lo_source_file + 9 );
      strcpy( lo_target_file, gsTargetDir );
      strcat( lo_target_file, "/" );
      strcat( lo_target_file, lo_source_file );   
      if (giVerbose>2) printf("- found INCOMING: %s\n", lo_source_file );
      if ( 0 == file_exists(lo_target_file) ) {
        if (giVerbose>0) {
	  printWARNING("- non sovrascrivo: %s\n", lo_target_file);
          normal_color();
	}
        goto_next_header( gzfd, lo_source_size );
	return 0;
      }
    }
    return process_file();    
      
  // Elabora il blocco come dati
  } else {

    // Controlla se c'e' un handle del file aperto
    if ( lo_target_fd != NULL ) {
      // Controlla se e' l'ultimo blocco del file
      if ( lo_target_tocopy > TAR_BLOCKSIZE ) {
        // Aggiunge blocco dati in coda al file
        //printf("untar: aggiunta blocco dati\n");
        if ( fwrite( &gzbuffer, sizeof(char), TAR_BLOCKSIZE, lo_target_fd ) != TAR_BLOCKSIZE ) {
          // In caso di errore non lascia il file monco sul FS
          printWARNING( "untar: error writing %s skipping...\n", lo_target_file );
          normal_color();
          fclose(lo_target_fd);
          unlink(lo_target_file);
          lo_target_fd = NULL;
          gzheader = 1;
	  sleep(1);
        }
        // Diminuisce in numero di byte da copiare
        lo_target_tocopy -= TAR_BLOCKSIZE;
        //printf("untar: rimangono byte %li\n", lo_target_tocopy );
      } else {
        //printf("untar: aggiunta ultimo blocco dati\n");
        if (lo_target_tocopy > 0 ) {
          if ( fwrite( &gzbuffer, sizeof(char), lo_target_tocopy, lo_target_fd ) != lo_target_tocopy ) {
            printWARNING("untar: error writing last Block of %s skipping...\n", lo_target_file );
	    normal_color();
            fclose(lo_target_fd);
            unlink(lo_target_file);
            lo_target_fd = NULL;
            gzheader = 1;
          }
        }
        fclose( lo_target_fd );
        //file_settime( lo_target_file, lo_source_time );
        lo_target_fd = NULL;
        gzheader = 1;
        //printf("- scrittura di %s completata\n", lo_target_file);
      }
    } else {
      printERROR("untar: blocco dati nr. %li ma nessun descrittore di file apero !\n", gzblock );
      normal_color();
      sleep(1);
      return ERROR_IO; 
    }
  }
  return 0;
}

// ================================================================
// *****                        M  A  I  N                    *****
// ================================================================

int untar(TTarOptions *lpopt) {
  int     gzerror = 0;		// Default return value
  int     gzbyte  = -1;		// Nr. of readed bytes from the zipped stream
  double  gzbar   = 0;
  
  // Check option structure
  if (lpopt->sourcefile == NULL) {
    printERROR("untar: sourcefile == NULL !\n");
    normal_color();
    return( ERROR_PARAMETER );
  }
  if (lpopt->targetdir == NULL) {
    printERROR("untar: targetdir == NULL !\n");
    normal_color();
    return( ERROR_PARAMETER );
  }

  gsRootDir   = lpopt->targetdir;
  gsTargetDir = lpopt->prefixdir;
  giMode    = lpopt->mode;
  giAction  = lpopt->action;
  giVerbose = lpopt->verbose;
  gsTime    = time(NULL);
  
  // Open .tar.gz file
  gzfd = gzopen( lpopt->sourcefile, "rb" );
  if (gzfd == NULL) {
    printERROR( "untar: gzopen, impossibile aprire il file compresso %s\n", lpopt->sourcefile );
    normal_color();
    sleep(1);
    return ERROR_IO;
  }

  // Set mask to 0700
  umask( S_IRWXG | S_IRWXO );

  // Make root dir if it does not exists
  if ( 0 != file_exists(gsTargetDir) ) {
    if (giVerbose>0) printf("> creating temp directory: %s \n", gsTargetDir );
    if ( 0 != makedir( gsTargetDir, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH ) ) {
      printWARNING("! unable to create temp dir: %s \n", gsTargetDir );
      normal_color();
      return ERROR_IO;
    }
  }
  
  // Read Data Blocks
  gzblock = -1;
  gzheader = 1;
  gzerror = 0;
  gzcompleted = 0;
  while ((gzerror == 0) && (gzcompleted == 0)) {

    // Legge un blocco di dati non compressi
    gzblock++;
    gzbyte = gzread( gzfd, (char*)(&gzbuffer), TAR_BLOCKSIZE );

    // PROGRESS BAR
    if (gzblock % 10 == 0) {
      if (lpopt->maxsize > 0) {
        gzbar  = (double)gztell(gzfd);
        gzbar *= 80;
        gzbar /= (double)lpopt->maxsize;
        fprintf(stdout, "> %4.1f %%\r", gzbar );
        fflush(stdout);
      }
    }
    
    // Controlla la corretta decompressione dei dati
    if ( gzbyte < 0 ) {
      printWARNING( "untar: errore (%i) nella zlib al blocco %li\n", gzbyte, gzblock );
      normal_color();
      sleep(1);
      break;
    } else
    if ( gzbyte == 0 ) {
      if (giVerbose>0) {
	printf("- decompressione archivio completata, %li blocchi.\n", gzblock);
      }
      break;
    } else
    if ( gzbyte != TAR_BLOCKSIZE ) {
      printWARNING( "untar: letti solo %i byte dal blocco %li\n", gzbyte, gzblock );
      normal_color();
      sleep(1);
      gzerror = ERROR_IO;
      break;
    }

    // Elabora il blocco letto
    gzerror = process_block();
  }
  
  // END untar
  if ( Z_OK != gzclose( gzfd ) ) {
    printERROR( "untar: fallito gzclose()\n");
    normal_color();
    sleep(1);
    return ERROR_IO;
  }

  // FlushCache
  if (gzerror == 0) {
    gzerror = FlushCache(lpopt);
  }

  // Exit
  return gzerror;
}
